// Utilities
#import "utils.typ" as utils
// Style presets
#import "style.typ" as styling
// ACM CCS generation
#import "acm-css.typ": apply-acm-ccs
// Import the title page
#import "title-page.typ": title-page
// Import the outline
#import "outline.typ": custom-outline, in-outline-state
// Import todo default state
#import "todo.typ": _show-notes, todo-list
// Import word counting
#import "@preview/wordometer:0.1.4": word-count, total-words

// Used for license iconography
#import "@preview/ccicons:1.0.0": ccicon

/// Apply the template to the main document
///
/// #example(`#show: style.with(...)`)
///
/// - document-type (string): idk
///
/// - title (string): The document Title
///
/// - authors (array): array of authers (dictionaries) with the following keys:
///   - first-name (string): The first name
///   - last-name (string): The last name
///
/// - drafting (bool): If the document is in an unfinished state (show todos etc.)
///
/// - doc (content): Apply the template to this document
///
/// TODO: update
///
/// -> content
#let template(
  document-type: "seminar",
  title: none,
  subtitle: none,
  running-title: auto,
  authors: none,
  supervisors: none,
  institution: (
    name: "Philipps-Universität Marburg",
    faculty: "Faculty of Mathematics and Computer Science",
    work-group: "Programming Languages and Tools Group",
    address: [
      Hans-Meerwein-Str. 6 \
      35043 Marburg \
      Deutschland
    ],
  ),
  abstract: lorem(100),
  appendix: none,
  bibliography-path: none,
  date: datetime.today(),
  acm-ccs: none,
  keywords: (),
  license: "by-nc-sa",
  drafting: true,
  doc,
) = {
  // Generate and set style
  let style = styling.default-style(document-type)
  if not drafting {
    style.todo.enable = false
  }
  styling.global-style.update(style)

  // -- Show & Set rules --

  show: word-count

  set page(
    paper: "a4",
    margin: (left: 3.15cm, right: 3.15cm, top: 3.5cm, bottom: 2.5cm),
    numbering: (num, ..rest) => context {
      let skipped_pages = 2
      if num > skipped_pages {
        text(
          ..if not in-outline-state.get() {
            style.heading-text-args
          },
          size: 10pt,
          weight: "medium",
          str(num - skipped_pages),
        )
      }
    },
    header: context {
      set text(..style.heading-text-args, weight: "medium", size: 0.9em)
      if calc.odd(counter(page).get().first()) {
        set align(right)

        text(
          authors
            .map(a => a.at("running-name", default: a.name))
            .join(", ", last: " and "),
        )
      } else if counter(page).get().first() != 2 {
        set align(left)

        if running-title == auto {
          title
        } else {
          running-title
        }
      }
    },
    footer-descent: 50%,
  )

  set document(
    title: title,
    author: authors.map(a => a.name),
    keywords: keywords,
  )

  show heading.where(level: 1): it => {
    block({
      {
        set box(outset: (y: 0.3em), fill: style.accent-color)
        if it.numbering == none {
          box(width: 1.43em, height: 0.48em, baseline: -0.05em)
        } else {
          box(
            inset: (x: 0.5em),
            fill: style.accent-color,
            text(
              top-edge: "bounds",
              bottom-edge: "bounds",
              numbering(it.numbering, ..counter(heading).get()),
              fill: white,
            ),
          )
        }
      }
      h(1em)
      it.body
    })
  }
  show heading: set text(
    ..style.heading-text-args,
    size: 0.9 * style.heading-text-args.at("size"),
  )
  show heading.where(level: 1): set text(..style.heading-text-args)
  show heading: set block(above: 2em, below: 2em)
  show heading.where(level: 1): set block(above: 3.3em, below: 1.9em)

  show raw: set text(..style.raw-text-args)
  show raw.where(block: true): set block(
    width: 100%,
    fill: style.code-background,
    inset: (x: 0.35em, y: 0.65em),
  )
  show raw.where(block: true): it => {
    show raw.line: it => [#box(
        place(
          center + bottom,
          [#text(
              fill: style.code-background.darken(20%),
              str(it.number),
            )],
          dx: -1em,
        ),
      ) #h(0em, weak: true) #it]
    it
  }

  show selector.or(
    ..(raw, table).map(kind => figure.where(kind: kind)),
  ): set figure.caption(position: top)
  show figure: set block(spacing: 1.5em)
  show figure.caption: it => {
    set align(left)

    block(
      width: 100%,
      {
        box(width: 0.6em, height: 0.6em, fill: style.accent-color)
        h(0.5em, weak: true)
        text(
          ..style.heading-text-args,
          size: 0.9em,
          [
            #it.supplement
            #context numbering(it.numbering, ..it.counter.get())
          ],
        )
        h(1em, weak: true)
        text(..style.normal-text-args, it.body, size: 0.9em)
      },
    )
  }

  show list: set block(spacing: 1em)
  set list(
    spacing: 0.75em,
    indent: 0.2em,
    marker: (
      box(
        width: 0.5em,
        height: 0.75em,
        align(horizon + right, box(width: 0.25em, height: 0.25em, fill: black)),
      ),
      [--],
    ),
  )

  show link: it => context {
    set text(
      ..style.heading-text-args,
      weight: "regular",
      size: 0.9em,
    ) if not in-outline-state.get()

    it
  }

  set footnote(numbering: (..nums) => super(size: 1em, numbering("1", ..nums)))
  set footnote.entry(
    separator: line(length: 2cm, stroke: 0.2pt),
    clearance: 1em,
  )

  set text(..style.normal-text-args)

  set par(
    justify: true,
    first-line-indent: 1.2em,
    spacing: 0.65em,
    leading: 0.6em,
  )

  if drafting {
    set heading(outlined: false)
    show heading: it => it.body

    page(numbering: "i")[
      #align(center, text(size: 3em, weight: "bold")[
        #total-words total words
      ])
      
      = Todo list

      #todo-list()
    ]
    counter(page).update(1)
  }

  title-page(
    subtitle: subtitle,
    institution,
    title,
    authors.map(a => a.name),
    supervisors,
    date,
    license,
    style,
  )

  // -- Document contnet --
  [
    #set heading(outlined: false)

    = Institution

    #institution.name \
    #institution.faculty \
    #institution.work-group \
    #institution.address

    = License

    This work is licensed under a Creative Commons #((sym.quote.single,) * 4).join() license.
    #h(1fr) #ccicon(license, format: "badge", link: true, scale: 3.3)

    #if acm-ccs != none {
      apply-acm-ccs(acm-ccs)
    }

    = Keywords

    #keywords.join(", ")

    = Abstract

    #abstract

    #custom-outline(style)
  ]

  show (1, 2, 3)
    .map(i => heading.where(level: i))
    .reduce(selector.or): set heading(numbering: "1.1")

  doc

  if bibliography-path != none [
    = References

    #bibliography(
      title: none,
      bibliography-path,
      style: "association-for-computing-machinery",
    )
  ]

  // Appendix
  if appendix != none {
    pagebreak(weak: true)
    set page(header: none)
    counter(heading).update(0)
    show heading: set heading(numbering: "A")

    text(..style.heading-text-args, size: 1.75em)[Appendix]
    v(-0.25em)
    line(length: 100%, stroke: style.accent-color + 0.15cm)

    appendix
  }
}
