// Used to generate the license badge
#import "@preview/ccicons:1.0.0": ccicon

/// Create the title page
///
/// - institution (dictionary):
///
///   A dictionary containing the following entries:
///
///   - name: The name of the institution
///   - faculty: The faculty of the institution
/// 
/// - title (string, content): The document title 
///
/// - subtitle (content, string): An optional subtitle
///
/// - authors (array): The authors names
///
/// - supervisors (dictionary):
///
///   A dictionary containing the following entries:
///
///   - role: The role of the supervisor
///   - info: The info of the supervisor
///
/// - date (datetime): The date of creation
///
/// - license (string): The license string for example `"by-nc-sa"`
///
/// - style (dictionary): the style, see `style.typ`
///
/// -> page
#let title-page(
  institution,
  title,
  subtitle: none,
  authors,
  supervisors,
  date,
  license,
  style,
) = {
  set par(justify: false)

  // Institution name and faculty
  let institution-title = {
    set align(right)
    set text(..style.heading-text-args)

    stack(
      spacing: 2em,
      text(
        size: 2.1em,
        institution.name,
      ),
      text(
        size: 1.2em,
        institution.faculty,
      ),
    )
  }

  // Title and subtitle
  let title = {
    set align(center)
    set text(..style.heading-text-args)
    set par(leading: 0.9em)
    stack(
      spacing: 1.5em,
      text(
        size: 1.8em,
        title,
      ),
      if subtitle != none {
        text(size: 1.5em, subtitle)
      },
    )
  }

  // Authors, document type and date
  let document-info = {
    set align(center)
    set text(..style.heading-text-args)

    stack(
      text(
        size: 1.3em,
        authors.join(
          ", ",
          last: ", and ",
        ),
      ),
      v(3em),
      text(size: 1.7em, fill: style.accent-color, style.document-description),
      v(1.8em),
      text(
        date.display("[month repr:long] [day], [year]"),
      ),
    )
  }

  // License icon and supervisors info
  let footer = align(
    bottom,
    {
      ccicon(license, format: "badge", link: true, scale: 3.25)
      h(1fr)
      box(
        outset: (left: 1cm),
        inset: (top: 0.75em, bottom: 1.2em),
        stroke: (top: 2pt, bottom: 2pt),
        [
          #set align(top + right)
          #set text(..style.heading-text-args, size: 0.9em)
          #set par(leading: 0.75em)
          #supervisors.map(s => stack(
            spacing: 0.75em,
            s.role + ":",
            text(s.info, weight: "regular"),
          )).join(v(1em))
        ],
      )
    },
  )

  // Put it all together
  page(
    header: none,
    footer: none,
    margin: (x: 0.5cm, y: 0cm),
    grid(
      columns: 2,
      {
        set image(height: 100%, fit: "stretch")
        style.title-image
      },
      box(
        inset: (top: 1.5cm, bottom: 1cm, x: 0.5cm),
        stack(
          spacing: 1fr,
          institution-title,
          title,
          document-info,
          footer,
        ),
      ),
    ),
  )
}
