// Import global style
#import "style.typ": global-style
// Import in-outline-state for todo list
#import "outline.typ": in-outline-state

/// Collect show notes into margins, and iterate through the pages to display them.
/// Intended to be used in a show rule e.g. `show: _show-notes`
///
/// TODO: restructure this (doesnt have to work like a show rule)
///
/// - body (content): The document
///
/// -> content
#let _show-notes(body) = context {
  // Keep track of pages with their margins drawn so we don't draw any twice
  let finished-margin-pages = state("finished-margin-pages", ())
  // Get style defaults
  let todo-defaults = global-style.get().todo

  // Collect all margin notes and store them according to their page
  let final-margin-notes = (left: (:), right: (:))
  for margin-note in query(<margin-note-anchor>) {
    let page-str = str(margin-note.location().page())
    let side = margin-note.value.side

    let page-notes = final-margin-notes.at(side).at(page-str, default: ())
    page-notes.push((
      note: margin-note.value.note,
      pos: margin-note.location().position(),
      color: margin-note.value.color,
    ))
    final-margin-notes.at(side).insert(page-str, page-notes)
  }

  /// Generate a path from the margin anchor to the margin note
  ///
  /// - side (alignment): the side of the note (left or right)
  ///
  /// - height (length): the height of the note
  ///
  /// - anchor (dictionary): the margin note anchor position (x, y)
  ///
  /// - color (color): the color of the path
  ///
  /// -> content
  let generate-margin-path(side, height, anchor, color) = {
    let start-pos = if side == left {
      (x: page.margin.left - todo-defaults.margin-inset.inner, y: height)
    } else {
      (
        x: page.width - page.margin.right + todo-defaults.margin-inset.inner,
        y: height,
      )
    }
    let end-pos = anchor
    let dy = end-pos.y - start-pos.y
    let dx = end-pos.x - start-pos.x
    let dir = if side == left {
      1
    } else {
      -1
    }

    let offset = here().position()
    place(
      dx: start-pos.x - offset.x,
      dy: start-pos.y - offset.y,
      curve(
        stroke: color + todo-defaults.line-width,
        curve.line((0cm, 0cm)),
        curve.line((dir * todo-defaults.margin-inset.inner / 2, 0cm)),
        curve.line((dir * todo-defaults.margin-inset.inner / 2, dy)),
        curve.line((dx, dy)),
      ),
    )
  }

  /// Generate one margin by collecting notes, stacking them and drawing their paths
  ///
  /// - side (alignment): The margin side (left or right)
  ///
  /// - notes (list): the margin notes
  ///
  /// -> content
  let generate-margin(side, notes) = {
    show rect: set block(above: 0cm, below: 0cm)

    let notes = notes.sorted(key: d => d.pos.y)
    let result = ()
    let margin-box = block.with(
      inset: if side == left {
        (
          left: todo-defaults.margin-inset.outer,
          right: todo-defaults.margin-inset.inner,
        )
      } else {
        (
          left: todo-defaults.margin-inset.inner,
          right: todo-defaults.margin-inset.outer,
        )
      },
      // WARNING: This is brittle and breaks if left and right margin are not set
      width: if side == left {
        page.margin.left
      } else {
        page.margin.right
      },
      height: auto,
    )

    // Generate the margin box (pushing spaces and notes as necessary)
    context {
      let result = result
      let height = 0cm
      for note in notes {
        // The difference between current note stack height and the preferred note height
        let difference = measure(v(note.pos.y - 1em - height)).height
        // Keep notes at least some distance apart
        let space = calc.max(difference, 0.25cm)
        // Measure the note height
        let note-height = measure(margin-box(note.note)).height
        result.push(v(space))
        result.push(note.note)
        // Update the used height
        height += space + note-height
        generate-margin-path(
          side,
          height - note-height / 2,
          note.pos,
          note.color,
        )
      }

      margin-box(result.join())
    }
  }

  // Generate the margins for each page
  final-margin-notes = (
    left: final-margin-notes
      .left
      .pairs()
      .map(((page, notes)) => (page, generate-margin(left, notes)))
      .to-dict(),
    right: final-margin-notes
      .right
      .pairs()
      .map(((page, notes)) => (page, generate-margin(right, notes)))
      .to-dict(),
  )

  // Iterate through the note anchors and draw margins at their pages
  // NOTE: this is somewhat of a workaround, since you can't use show rules directly on pages
  show <margin-note-anchor>: it => {
    let anchor-page = it.location().page()
    let pos = it.location().position()

    if not anchor-page in finished-margin-pages.get() {
      // Keep track of finished pages, see the limitation above
      finished-margin-pages.update(pages => {
        pages.push(page)
        pages
      })

      // Place the left margin
      let left-margin = final-margin-notes.left.at(
        str(anchor-page),
        default: none,
      )
      if left-margin != none {
        place(
          top + left,
          dx: -pos.x,
          dy: -pos.y,
          left-margin,
        )
      }

      // Place the right margin
      let right-margin = final-margin-notes.right.at(
        str(anchor-page),
        default: none,
      )
      place(
        top + left,
        dx: -pos.x + page.width - page.margin.right,
        dy: -pos.y,
        right-margin,
      )
    }
  }

  body
}

/// Draw a single note
///
/// - body (content): the note body
///
/// - color (color): the background color
///
/// - defaults (dictionary): the default note style
///
/// - author (dictionary): author information, containing the following keys:
///   - name (string): the name of the author
///   - count (int): the index of the todo item
///
/// -> content
#let _note-rect(body, color, author: none, defaults) = {
  set rect(fill: color)
  set par(justify: false)
  set block(spacing: 0.5em)

  (defaults.rect)({
    if author != none {
      [*[#author.name#sub(str(author.count))]: *]
    }
    body
  })
}

/// Draw a single inline note
///
/// - body (content): the note body
///
/// - color (color): the background color
///
/// - defaults (dictionary): the default note style
///
/// - author (dictionary): author information
///
/// -> content
#let _inline-note(body, color, author: none, defaults) = {
  _note-rect(body, color, author: author, defaults)
}

/// Draw a margin note
///
/// - body (content): the note body
///
/// - color (color): the background color
///
/// - defaults (dictionary): the default note style
///
/// - side (alignment): the margin at which to draw the note (left or right)
///
/// - author (dictionary): author information
///
/// -> content
#let _margin-note(body, color, side, author: none, defaults) = {
  let side = if side == right {
    "right"
  } else if side == left {
    "left"
  } else {
    panic("Expected side alignment of left or right, got: " + side)
  }
  let note = _note-rect(body, color, author: author, defaults)
  {
    show: _show-notes
    box(
      place(
        dy: defaults.underline-offset,
        box[
          #metadata((
            note: note,
            side: side,
            color: color,
          ))<margin-note-anchor>
        ],
      ),
    )
  }
}

/// Generate a todo marker for use in the todo list
///
/// - description (content): The description in the todo list
///
/// - author (dictionary): The author and count
///
/// - type (string): The type of the element (either `"figure"` or `"note"`)
///
/// -> content
#let _todo-marker(description, author: none, type: "note") = {
  [#metadata((
      description: description,
      author: author,
      type: type,
    ))<todo-marker>]
}

/// Generate a todo note
///
/// - inline (bool): if the note should be inline or a mragin note
///
/// - side (alignment): the margin side to use (either left or right). Has no effect on inline notes
///
/// - color (color): the color used to fill the note background
///
/// - body (content): The note body and optional preceding highlighted elements
///                   (the last element is the note body)
///
/// -> content
#let todo(
  inline: false,
  side: right,
  color: auto,
  author: none,
  ..body,
) = context {
  let defaults = global-style.get().todo
  assert(defaults != none, message: "Todos only work when using the template")

  let color = if color == auto {
    defaults.color
  } else {
    color
  }

  assert(
    body.named().len() == 0,
    message: "Expected no additional named arguments",
  )
  if not defaults.enable {
    return
  }

  let c = counter(if author != none {
    author + "-todo-counter"
  } else {
    "default-todo-counter"
  })
  c.step()

  let bodies = body.pos()
  bodies
    .slice(0, -1)
    .map(b => underline(
      stroke: color + defaults.line-width,
      offset: defaults.underline-offset,
      highlight(b),
    ))
    .join()

  context {
    let author-num = c.get().first()
    let author = if author != none {
      (name: author, count: author-num)
    }

    text(
      size: 0.8em,
      if inline {
        _inline-note(
          bodies.at(-1),
          color,
          author: author,
          defaults,
        )
      } else {
        _margin-note(
          bodies.at(-1),
          color,
          side,
          author: author,
          defaults,
        )
      },
    )
    [#_todo-marker(bodies.at(-1), author: author)]
  }
}

/// Generate a "missing figure" box with a description
///
/// - description (content): The description for the figure
///
/// - width (length): The width of the box
///
/// - height (length): The height of the box
///
/// - author (string): The author of this todo
///
/// -> content
#let todo-missing-figure(
  description,
  width: 100%,
  height: 4cm,
  author: none,
) = {
  let c = counter(if author != none {
    author + "-todo-counter"
  } else {
    "default-todo-counter"
  })
  c.step()

  context {
    let author = if author != none {
      (
        name: author,
        count: c.get().first(),
      )
    }
    rect(
      fill: gray,
      height: height,
      width: 100%,
      inset: (left: 1.25em, top: 0.5em),
      {
        _todo-marker(description, author: author, type: "figure")
        align(
          left + horizon,
          stack(
            dir: ltr,
            spacing: 1.5em,
            box({
              scale(
                y: -100%,
                polygon.regular(
                  vertices: 3,
                  size: height,
                  fill: white,
                  stroke: (
                    paint: red.saturate(100%),
                    thickness: 10pt,
                    join: "round",
                  ),
                ),
              )
              place(center + horizon, [Missing \ figure], dy: -1.5em)
            }),
            text(
              size: 0.8em,
              align(
                left + horizon,
                {
                  if author != none {
                    [*[#author.name#sub(str(author.count))]: *]
                  }
                  description
                },
              ),
            ),
          ),
        )
      },
    )
  }
}

/// Generate an overview of all todo items.
///
/// -> content
#let todo-list() = {
  set block(spacing: 2em)
  set par(hanging-indent: 2em, justify: false)

  let dots = box(
    width: 1fr,
    align(
      right,
      {
        repeat(
          gap: 0.5em,
          justify: false,
        )[.]
      },
    ),
  )

  context grid(
    column-gutter: 2em,
    row-gutter: 0.75em,
    columns: 2,
    align: (left, right + bottom),
    ..query(selector(<todo-marker>))
      .map(e => (
        {
          link(
            e.location(),
            {
              if e.value.type == "figure" {
                [Figure: ]
              }
              if e.value.author != none {
                [*[#e.value.author.name#sub(str(e.value.author.count))]: *]
              }
              e.value.description
            },
          )
          dots
        },
        numbering(
          e.location().page-numbering(),
          ..counter(page).at(e.location()),
        ),
      ))
      .flatten()
  )
}
